@@ -1,21 +1,17 @@
 /*
-    Copyright 2005-2015 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2020 Intel Corporation
 
-    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
-    you can redistribute it and/or modify it under the terms of the GNU General Public License
-    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
-    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
-    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-    See  the GNU General Public License for more details.   You should have received a copy of
-    the  GNU General Public License along with Threading Building Blocks; if not, write to the
-    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
 
-    As a special exception,  you may use this file  as part of a free software library without
-    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
-    functions from this file, or you compile this file and link it with other files to produce
-    an executable,  this file does not by itself cause the resulting executable to be covered
-    by the GNU General Public License. This exception does not however invalidate any other
-    reasons why the executable file might be covered by the GNU General Public License.
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
 */
 
 #include "tbb/tbb_config.h"
@@ -39,15 +35,13 @@ static const char _pad[NFS_MaxLineSize - sizeof(int)] = {};
 
 //------------------------------------------------------------------------
 // governor data
-basic_tls<generic_scheduler*> governor::theTLS;
+basic_tls<uintptr_t> governor::theTLS;
 unsigned governor::DefaultNumberOfThreads;
+size_t governor::DefaultPageSize;
 rml::tbb_factory governor::theRMLServerFactory;
 bool governor::UsePrivateRML;
-const task_scheduler_init *governor::BlockingTSI;
-#if TBB_USE_ASSERT
-bool governor::IsBlockingTerminationInProgress;
-#endif
 bool governor::is_speculation_enabled;
+bool governor::is_rethrow_broken;
 
 //------------------------------------------------------------------------
 // market data
@@ -67,7 +61,7 @@ bool __TBB_InitOnce::InitializationDone;
 
 #if DO_ITT_NOTIFY
     static bool ITT_Present;
-    static bool ITT_InitializationDone;
+    static atomic<bool> ITT_InitializationDone;
 #endif
 
 #if !(_WIN32||_WIN64) || __TBB_SOURCE_DIRECTLY_INCLUDED
@@ -78,7 +72,7 @@ bool __TBB_InitOnce::InitializationDone;
 // generic_scheduler data
 
 //! Pointer to the scheduler factory function
-generic_scheduler* (*AllocateSchedulerPtr)( arena*, size_t index );
+generic_scheduler* (*AllocateSchedulerPtr)( market&, bool );
 
 #if __TBB_OLD_PRIMES_RNG
 //! Table of primes used by fast random-number generator (FastRandom).
@@ -127,7 +121,7 @@ void __TBB_InitOnce::add_ref() {
 
 void __TBB_InitOnce::remove_ref() {
     int k = --count;
-    __TBB_ASSERT(k>=0,"removed __TBB_InitOnce ref that was not added?"); 
+    __TBB_ASSERT(k>=0,"removed __TBB_InitOnce ref that was not added?");
     if( k==0 ) {
         governor::release_resources();
         ITT_FINI_ITTLIB();
@@ -146,9 +140,7 @@ void Scheduler_OneTimeInitialization ( bool itt_present );
 
 #if DO_ITT_NOTIFY
 
-#if __TBB_ITT_STRUCTURE_API
-
-static __itt_domain *fgt_domain = NULL;
+static __itt_domain *tbb_domains[ITT_NUM_DOMAINS] = {};
 
 struct resource_string {
     const char *str;
@@ -161,18 +153,22 @@ struct resource_string {
 #define TBB_STRING_RESOURCE( index_name, str ) { str, NULL },
 static resource_string strings_for_itt[] = {
     #include "tbb/internal/_tbb_strings.h"
-    { "num_resource_strings", NULL } 
+    { "num_resource_strings", NULL }
 };
 #undef TBB_STRING_RESOURCE
 
 static __itt_string_handle *ITT_get_string_handle(int idx) {
-    __TBB_ASSERT(idx >= 0, NULL);
-    return idx < NUM_STRINGS ? strings_for_itt[idx].itt_str_handle : NULL;
+    __TBB_ASSERT( idx >= 0 && idx < NUM_STRINGS, "string handle out of valid range");
+    return (idx >= 0 && idx < NUM_STRINGS) ? strings_for_itt[idx].itt_str_handle : NULL;
 }
 
 static void ITT_init_domains() {
-    fgt_domain = __itt_domain_create( _T("tbb.flow") );
-    fgt_domain->flags = 1;
+    tbb_domains[ITT_DOMAIN_MAIN] = __itt_domain_create( _T("tbb") );
+    tbb_domains[ITT_DOMAIN_MAIN]->flags = 1;
+    tbb_domains[ITT_DOMAIN_FLOW] = __itt_domain_create( _T("tbb.flow") );
+    tbb_domains[ITT_DOMAIN_FLOW]->flags = 1;
+    tbb_domains[ITT_DOMAIN_ALGO] = __itt_domain_create( _T("tbb.algorithm") );
+    tbb_domains[ITT_DOMAIN_ALGO]->flags = 1;
 }
 
 static void ITT_init_strings() {
@@ -190,16 +186,14 @@ static void ITT_init() {
     ITT_init_strings();
 }
 
-#endif // __TBB_ITT_STRUCTURE_API
-
 /** Thread-unsafe lazy one-time initialization of tools interop.
     Used by both dummy handlers and general TBB one-time initialization routine. **/
 void ITT_DoUnsafeOneTimeInitialization () {
+    // Double check ITT_InitializationDone is necessary because the first check 
+    // in ITT_DoOneTimeInitialization is not guarded with the __TBB_InitOnce lock.
     if ( !ITT_InitializationDone ) {
         ITT_Present = (__TBB_load_ittnotify()!=0);
-#if __TBB_ITT_STRUCTURE_API
         if (ITT_Present) ITT_init();
-#endif
         ITT_InitializationDone = true;
         ITT_SYNC_CREATE(&market::theMarketMutex, SyncType_GlobalLock, SyncObj_SchedulerInitialization);
     }
@@ -209,9 +203,11 @@ void ITT_DoUnsafeOneTimeInitialization () {
     Used by dummy handlers only. **/
 extern "C"
 void ITT_DoOneTimeInitialization() {
-    __TBB_InitOnce::lock();
-    ITT_DoUnsafeOneTimeInitialization();
-    __TBB_InitOnce::unlock();
+    if ( !ITT_InitializationDone ) {
+        __TBB_InitOnce::lock();
+        ITT_DoUnsafeOneTimeInitialization();
+        __TBB_InitOnce::unlock();
+    }
 }
 #endif /* DO_ITT_NOTIFY */
 
@@ -234,6 +230,8 @@ void DoOneTimeInitializations() {
         Scheduler_OneTimeInitialization( itt_present );
         // Force processor groups support detection
         governor::default_num_threads();
+        // Force OS regular page size detection
+        governor::default_page_size();
         // Dump version data
         governor::print_version_info();
         PrintExtraVersionInfo( "Tools support", itt_present ? "enabled" : "disabled" );
@@ -244,12 +242,16 @@ void DoOneTimeInitializations() {
 
 #if (_WIN32||_WIN64) && !__TBB_SOURCE_DIRECTLY_INCLUDED
 //! Windows "DllMain" that handles startup and shutdown of dynamic library.
-extern "C" bool WINAPI DllMain( HANDLE /*hinstDLL*/, DWORD reason, LPVOID /*lpvReserved*/ ) {
+extern "C" bool WINAPI DllMain( HANDLE /*hinstDLL*/, DWORD reason, LPVOID lpvReserved ) {
     switch( reason ) {
         case DLL_PROCESS_ATTACH:
             __TBB_InitOnce::add_ref();
             break;
         case DLL_PROCESS_DETACH:
+            // Since THREAD_DETACH is not called for the main thread, call auto-termination
+            // here as well - but not during process shutdown (due to risk of a deadlock).
+            if( lpvReserved==NULL ) // library unload
+                governor::terminate_auto_initialized_scheduler();
             __TBB_InitOnce::remove_ref();
             // It is assumed that InitializationDone is not set after DLL_PROCESS_DETACH,
             // and thus no race on InitializationDone is possible.
@@ -276,7 +278,7 @@ void* itt_load_pointer_with_acquire_v3( const void* src ) {
     ITT_NOTIFY(sync_acquired, const_cast<void*>(src));
     return result;
 }
-    
+
 #if DO_ITT_NOTIFY
 void call_itt_notify_v5(int t, void *ptr) {
     switch (t) {
@@ -290,14 +292,14 @@ void call_itt_notify_v5(int t, void *ptr) {
 void call_itt_notify_v5(int /*t*/, void* /*ptr*/) {}
 #endif
 
-#if __TBB_ITT_STRUCTURE_API
-
 #if DO_ITT_NOTIFY
-
 const __itt_id itt_null_id = {0, 0, 0};
 
 static inline __itt_domain* get_itt_domain( itt_domain_enum idx ) {
-    return ( idx == ITT_DOMAIN_FLOW ) ? fgt_domain : NULL;
+    if (tbb_domains[idx] == NULL) {
+        ITT_DoOneTimeInitialization();
+    }
+    return tbb_domains[idx];
 }
 
 static inline void itt_id_make(__itt_id *id, void* addr, unsigned long long extra) {
@@ -308,7 +310,7 @@ static inline void itt_id_create(const __itt_domain *domain, __itt_id id) {
     ITTNOTIFY_VOID_D1(id_create, domain, id);
 }
 
-void itt_make_task_group_v7( itt_domain_enum domain, void *group, unsigned long long group_extra, 
+void itt_make_task_group_v7( itt_domain_enum domain, void *group, unsigned long long group_extra,
                              void *parent, unsigned long long parent_extra, string_index name_index ) {
     if ( __itt_domain *d = get_itt_domain( domain ) ) {
         __itt_id group_id = itt_null_id;
@@ -323,7 +325,7 @@ void itt_make_task_group_v7( itt_domain_enum domain, void *group, unsigned long
     }
 }
 
-void itt_metadata_str_add_v7( itt_domain_enum domain, void *addr, unsigned long long addr_extra, 
+void itt_metadata_str_add_v7( itt_domain_enum domain, void *addr, unsigned long long addr_extra,
                               string_index key, const char *value ) {
     if ( __itt_domain *d = get_itt_domain( domain ) ) {
         __itt_id id = itt_null_id;
@@ -338,27 +340,45 @@ void itt_metadata_str_add_v7( itt_domain_enum domain, void *addr, unsigned long
     }
 }
 
-void itt_relation_add_v7( itt_domain_enum domain, void *addr0, unsigned long long addr0_extra, 
-                          itt_relation relation, void *addr1, unsigned long long addr1_extra ) {
+void itt_metadata_ptr_add_v11( itt_domain_enum domain, void *addr, unsigned long long addr_extra,
+                              string_index key, void *value ) {
     if ( __itt_domain *d = get_itt_domain( domain ) ) {
-        __itt_id id0 = itt_null_id; 
-        __itt_id id1 = itt_null_id;
-        itt_id_make( &id0, addr0, addr0_extra );
-        itt_id_make( &id1, addr1, addr1_extra );
-        ITTNOTIFY_VOID_D3(relation_add, d, id0, (__itt_relation)relation, id1); 
+        __itt_id id = itt_null_id;
+        itt_id_make( &id, addr, addr_extra );
+        __itt_string_handle *k = ITT_get_string_handle(key);
+#if __TBB_x86_32
+        ITTNOTIFY_VOID_D5(metadata_add, d, id, k, __itt_metadata_u32, 1, value);
+#else
+        ITTNOTIFY_VOID_D5(metadata_add, d, id, k, __itt_metadata_u64, 1, value);
+#endif 
     }
 }
 
-void itt_task_begin_v7( itt_domain_enum domain, void *task, unsigned long long task_extra, 
-                        void *parent, unsigned long long parent_extra, string_index /* name_index */ ) {
+
+void itt_relation_add_v7( itt_domain_enum domain, void *addr0, unsigned long long addr0_extra,
+                          itt_relation relation, void *addr1, unsigned long long addr1_extra ) {
+    if ( __itt_domain *d = get_itt_domain( domain ) ) {
+        __itt_id id0 = itt_null_id;
+        __itt_id id1 = itt_null_id;
+        itt_id_make( &id0, addr0, addr0_extra );
+        itt_id_make( &id1, addr1, addr1_extra );
+        ITTNOTIFY_VOID_D3(relation_add, d, id0, (__itt_relation)relation, id1);
+    }
+}
+
+void itt_task_begin_v7( itt_domain_enum domain, void *task, unsigned long long task_extra,
+                        void *parent, unsigned long long parent_extra, string_index name_index ) {
     if ( __itt_domain *d = get_itt_domain( domain ) ) {
         __itt_id task_id = itt_null_id;
         __itt_id parent_id = itt_null_id;
-        itt_id_make( &task_id, task, task_extra );
+        if ( task ) {
+            itt_id_make( &task_id, task, task_extra );
+        }
         if ( parent ) {
             itt_id_make( &parent_id, parent, parent_extra );
         }
-        ITTNOTIFY_VOID_D3(task_begin, d, task_id, parent_id, NULL );
+        __itt_string_handle *n = ITT_get_string_handle(name_index);
+        ITTNOTIFY_VOID_D3(task_begin, d, task_id, parent_id, n );
     }
 }
 
@@ -368,26 +388,53 @@ void itt_task_end_v7( itt_domain_enum domain ) {
     }
 }
 
+void itt_region_begin_v9( itt_domain_enum domain, void *region, unsigned long long region_extra,
+                          void *parent, unsigned long long parent_extra, string_index /* name_index */ ) {
+    if ( __itt_domain *d = get_itt_domain( domain ) ) {
+        __itt_id region_id = itt_null_id;
+        __itt_id parent_id = itt_null_id;
+        itt_id_make( &region_id, region, region_extra );
+        if ( parent ) {
+            itt_id_make( &parent_id, parent, parent_extra );
+        }
+        ITTNOTIFY_VOID_D3(region_begin, d, region_id, parent_id, NULL );
+    }
+}
+
+void itt_region_end_v9( itt_domain_enum domain, void *region, unsigned long long region_extra ) {
+    if ( __itt_domain *d = get_itt_domain( domain ) ) {
+        __itt_id region_id = itt_null_id;
+        itt_id_make( &region_id, region, region_extra );
+        ITTNOTIFY_VOID_D1( region_end, d, region_id );
+    }
+}
+
 #else // DO_ITT_NOTIFY
 
-void itt_make_task_group_v7( itt_domain_enum domain, void *group, unsigned long long group_extra, 
-                             void *parent, unsigned long long parent_extra, string_index name_index ) { }
+void itt_make_task_group_v7( itt_domain_enum /*domain*/, void* /*group*/, unsigned long long /*group_extra*/,
+                             void* /*parent*/, unsigned long long /*parent_extra*/, string_index /*name_index*/ ) { }
 
-void itt_metadata_str_add_v7( itt_domain_enum domain, void *addr, unsigned long long addr_extra, 
-                              string_index key, const char *value ) { }
+void itt_metadata_str_add_v7( itt_domain_enum /*domain*/, void* /*addr*/, unsigned long long /*addr_extra*/,
+                              string_index /*key*/, const char* /*value*/ ) { }
 
-void itt_relation_add_v7( itt_domain_enum domain, void *addr0, unsigned long long addr0_extra, 
-                          itt_relation relation, void *addr1, unsigned long long addr1_extra ) { }
+void itt_relation_add_v7( itt_domain_enum /*domain*/, void* /*addr0*/, unsigned long long /*addr0_extra*/,
+                          itt_relation /*relation*/, void* /*addr1*/, unsigned long long /*addr1_extra*/ ) { }
 
-void itt_task_begin_v7( itt_domain_enum domain, void *task, unsigned long long task_extra, 
-                        void * /*parent*/, unsigned long long /* parent_extra */, string_index /* name_index */ ) { }
+void itt_metadata_ptr_add_v11( itt_domain_enum /*domain*/, void * /*addr*/, unsigned long long /*addr_extra*/,
+                              string_index /*key*/, void * /*value*/ ) {}
 
-void itt_task_end_v7( itt_domain_enum domain ) { }
+void itt_task_begin_v7( itt_domain_enum /*domain*/, void* /*task*/, unsigned long long /*task_extra*/,
+                        void* /*parent*/, unsigned long long /*parent_extra*/, string_index /*name_index*/ ) { }
+
+void itt_task_end_v7( itt_domain_enum /*domain*/ ) { }
+
+void itt_region_begin_v9( itt_domain_enum /*domain*/, void* /*region*/, unsigned long long /*region_extra*/,
+                          void* /*parent*/, unsigned long long /*parent_extra*/, string_index /*name_index*/ ) { }
+
+void itt_region_end_v9( itt_domain_enum /*domain*/, void* /*region*/, unsigned long long /*region_extra*/ ) { }
 
 #endif // DO_ITT_NOTIFY
 
-#endif // __TBB_ITT_STRUCTURE_API
-
 void* itt_load_pointer_v3( const void* src ) {
     //TODO: replace this with __TBB_load_relaxed
     void* result = *static_cast<void*const*>(src);
@@ -399,7 +446,6 @@ void itt_set_sync_name_v3( void* obj, const tchar* name) {
     suppress_unused_warning(obj, name);
 }
 
-
 class control_storage {
     friend class tbb::interface9::global_control;
 protected:
@@ -418,19 +464,18 @@ protected:
 };
 
 class allowed_parallelism_control : public padded<control_storage> {
-    virtual size_t default_value() const {
-        // current implementation can't have effective active value below 2
-        return max(2U, governor::default_num_threads());
+    virtual size_t default_value() const __TBB_override {
+        return max(1U, governor::default_num_threads());
     }
-    virtual bool is_first_arg_preferred(size_t a, size_t b) const {
+    virtual bool is_first_arg_preferred(size_t a, size_t b) const __TBB_override {
         return a<b; // prefer min allowed parallelism
     }
-    virtual void apply_active() const {
-        __TBB_ASSERT( my_active_value>=2, NULL );
+    virtual void apply_active() const __TBB_override {
+        __TBB_ASSERT( my_active_value>=1, NULL );
         // -1 to take master into account
         market::set_active_num_workers( my_active_value-1 );
     }
-    virtual size_t active_value() const {
+    virtual size_t active_value() const __TBB_override {
 /* Reading of my_active_value is not synchronized with possible updating
    of my_head by other thread. It's ok, as value of my_active_value became
    not invalid, just obsolete. */
@@ -449,12 +494,12 @@ public:
 };
 
 class stack_size_control : public padded<control_storage> {
-    virtual size_t default_value() const {
+    virtual size_t default_value() const __TBB_override {
         return tbb::internal::ThreadStackSize;
     }
-    virtual void apply_active() const {
-#if __TBB_WIN8UI_SUPPORT
-        __TBB_ASSERT( false, "For Windows Store* apps we must not set stack size" );
+    virtual void apply_active() const __TBB_override {
+#if __TBB_WIN8UI_SUPPORT && (_WIN32_WINNT < 0x0A00)
+        __TBB_ASSERT( false, "For Windows 8 Store* apps we must not set stack size" );
 #endif
     }
 };
