@@ -1,21 +1,17 @@
 /*
-    Copyright 2005-2015 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2020 Intel Corporation
 
-    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
-    you can redistribute it and/or modify it under the terms of the GNU General Public License
-    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
-    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
-    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-    See  the GNU General Public License for more details.   You should have received a copy of
-    the  GNU General Public License along with Threading Building Blocks; if not, write to the
-    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
 
-    As a special exception,  you may use this file  as part of a free software library without
-    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
-    functions from this file, or you compile this file and link it with other files to produce
-    an executable,  this file does not by itself cause the resulting executable to be covered
-    by the GNU General Public License. This exception does not however invalidate any other
-    reasons why the executable file might be covered by the GNU General Public License.
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
 */
 
 // Declarations for rock-bottom simple test harness.
@@ -60,16 +56,10 @@ int TestMain ();
     #include <ucontext.h>
 #else /* !__SUNPRO_CC */
     #include <cstdlib>
-#if !TBB_USE_EXCEPTIONS && _MSC_VER
-    // Suppress "C++ exception handler used, but unwind semantics are not enabled" warning in STL headers
-    #pragma warning (push)
-    #pragma warning (disable: 4530)
-#endif
     #include <cstring>
-#if !TBB_USE_EXCEPTIONS && _MSC_VER
-    #pragma warning (pop)
-#endif
 #endif /* !__SUNPRO_CC */
+#include <cerrno>
+#include <cctype>
 
 #include <new>
 
@@ -83,13 +73,13 @@ int TestMain ();
 #if _WIN32||_WIN64
     #include "tbb/machine/windows_api.h"
     #if _WIN32_WINNT > 0x0501 && _MSC_VER && !_M_ARM
+        // Suppress "typedef ignored ... when no variable is declared" warning by vc14
+        #pragma warning (push)
+        #pragma warning (disable: 4091)
         #include <dbghelp.h>
+        #pragma warning (pop)
         #pragma comment (lib, "dbghelp.lib")
     #endif
-    #if _XBOX
-        #undef HARNESS_NO_PARSE_COMMAND_LINE
-        #define HARNESS_NO_PARSE_COMMAND_LINE 1
-    #endif
     #if __TBB_WIN8UI_SUPPORT
         #include <thread>
     #endif
@@ -112,11 +102,51 @@ int TestMain ();
     #define BACKTRACE_FUNCTION_AVAILABLE 1
 #endif
 
+namespace Harness {
+    class NativeMutex {
+#if _WIN32||_WIN64
+        CRITICAL_SECTION my_critical_section;
+      public:
+        NativeMutex() {
+            InitializeCriticalSectionEx(&my_critical_section, 4000, 0);
+        }
+        void lock() {
+            EnterCriticalSection(&my_critical_section);
+        }
+        void unlock() {
+            LeaveCriticalSection(&my_critical_section);
+        }
+        ~NativeMutex() {
+            DeleteCriticalSection(&my_critical_section);
+        }
+#else
+        pthread_mutex_t m_mutex;
+    public:
+        NativeMutex() {
+             pthread_mutex_init(&m_mutex, NULL);
+        }
+        void lock() {
+            pthread_mutex_lock(&m_mutex);
+        }
+        void unlock() {
+            pthread_mutex_unlock(&m_mutex);
+        }
+        ~NativeMutex() {
+            pthread_mutex_destroy(&m_mutex);
+        }
+#endif
+    };
+    namespace internal {
+        static NativeMutex print_stack_mutex;
+    }
+}
+
 #include "harness_runtime_loader.h"
 #include "harness_report.h"
 
 //! Prints current call stack
 void print_call_stack() {
+    Harness::internal::print_stack_mutex.lock();
     fflush(stdout); fflush(stderr);
     #if BACKTRACE_FUNCTION_AVAILABLE
         const int sz = 100; // max number of frames to capture
@@ -127,7 +157,7 @@ void print_call_stack() {
     #elif __SUNPRO_CC
         REPORT("Call stack info:\n");
         printstack(fileno(stdout));
-    #elif _WIN32_WINNT > 0x0501 && _MSC_VER && !__TBB_WIN8UI_SUPPORT
+    #elif _WIN32_WINNT > 0x0501 && _MSC_VER>=1500 && !__TBB_WIN8UI_SUPPORT
         const int sz = 62; // XP limitation for number of frames
         void *buff[sz];
         int n = CaptureStackBackTrace(0, sz, buff, NULL);
@@ -147,6 +177,7 @@ void print_call_stack() {
             REPORT("[%d] %016I64X+%04I64X: %s\n", i, sym.Address, offset, sym.Name); //TODO: print module name
         }
     #endif /*BACKTRACE_FUNCTION_AVAILABLE*/
+    Harness::internal::print_stack_mutex.unlock();
 }
 
 #if !HARNESS_NO_ASSERT
@@ -277,7 +308,7 @@ static int MaxThread = HARNESS_DEFAULT_MAX_THREADS;
     A single number m is interpreted as if written m:m.
     The numbers must be non-negative.
     Clients often treat the value 0 as "run sequentially." */
-static void ParseCommandLine( int argc, char* argv[] ) {
+inline void ParseCommandLine( int argc, char* argv[] ) {
     if( !argc ) REPORT("Command line with 0 arguments\n");
     int i = 1;
     if( i<argc ) {
@@ -391,8 +422,9 @@ int main(int argc, char* argv[]) {
         res = TestMain();
 #if __TBB_MIC_OFFLOAD && __MIC__
         // It is recommended not to use the __MIC__ macro directly in the offload block but it is Ok here
-        // since it is not lead to an unexpected difference between host and target compilation phases.
-        // We need to flush internals COI buffers to order output from the offload part before the host part.
+        // since it does not lead to an unexpected difference between host and target compilation phases.
+        // We need to flush internal Intel(R) Coprocessor Offload Infrastructure (Intel(R) COI) buffers
+        // to order output from the offload part before the host part.
         // Also it is work-around for the issue with missed output.
         COIProcessProxyFlush();
 #endif
@@ -413,15 +445,31 @@ int main(int argc, char* argv[]) {
 
 #endif /* !HARNESS_CUSTOM_MAIN */
 
+#if __TBB_DEFAULTED_AND_DELETED_FUNC_PRESENT
+
+//! Base class for types that should not be assigned.
+class NoAssign {
+public:
+    void operator=( const NoAssign& ) = delete;
+    NoAssign( const NoAssign& ) = default;
+    NoAssign() = default;
+};
+
+//! Base class for types that should not be copied or assigned.
+class NoCopy: NoAssign {
+public:
+    NoCopy( const NoCopy& ) = delete;
+    NoCopy() = default;
+};
+
+#else /*__TBB_DEFAULTED_AND_DELETED_FUNC_PRESENT*/
+
 //! Base class for prohibiting compiler-generated operator=
 class NoAssign {
     //! Assignment not allowed
     void operator=( const NoAssign& );
 public:
-#if __GNUC__
-    //! Explicitly define default construction, because otherwise gcc issues gratuitous warning.
-    NoAssign() {}
-#endif /* __GNUC__ */
+    NoAssign() {} // explicitly defined to prevent gratuitous warnings
 };
 
 //! Base class for prohibiting compiler-generated copy constructor or operator=
@@ -432,6 +480,47 @@ public:
     NoCopy() {}
 };
 
+#endif /*__TBB_DEFAULTED_AND_DELETED_FUNC_PRESENT*/
+
+#if __TBB_CPP11_RVALUE_REF_PRESENT
+#include <utility>
+
+//! Base class for objects which support move ctors
+class Movable {
+public:
+    Movable() : alive(true) {}
+    void Reset() { alive = true; }
+    Movable(Movable&& other) {
+        ASSERT(other.alive, "Moving from a dead object");
+        alive = true;
+        other.alive = false;
+    }
+    Movable& operator=(Movable&& other) {
+        ASSERT(alive, "Assignment to a dead object");
+        ASSERT(other.alive, "Assignment of a dead object");
+        other.alive = false;
+        return *this;
+    }
+    Movable& operator=(const Movable& other) {
+        ASSERT(alive, "Assignment to a dead object");
+        ASSERT(other.alive, "Assignment of a dead object");
+        return *this;
+    }
+    Movable(const Movable& other) {
+        ASSERT(other.alive, "Const reference to a dead object");
+        alive = true;
+    }
+    ~Movable() { alive = false; }
+    volatile bool alive;
+};
+
+class MoveOnly : Movable, NoCopy {
+public:
+    MoveOnly() : Movable() {}
+    MoveOnly(MoveOnly&& other) : Movable( std::move(other) ) {}
+};
+#endif /* __TBB_CPP11_RVALUE_REF_PRESENT */
+
 #if HARNESS_TBBMALLOC_THREAD_SHUTDOWN && __TBB_SOURCE_DIRECTLY_INCLUDED && (_WIN32||_WIN64)
 #include "../tbbmalloc/tbbmalloc_internal_api.h"
 #endif
@@ -470,8 +559,6 @@ public:
         // launched by make, the default stack size is set to the hard limit, and
         // calls to pthread_create fail with out-of-memory error.
         // Therefore we set the stack size explicitly (as for TBB worker threads).
-// TODO: make a single definition of MByte used by all tests.
-        const size_t MByte = 1024*1024;
 #if !defined(HARNESS_THREAD_STACK_SIZE)
 #if __i386__||__i386||__arm__
         const size_t stack_size = 1*MByte;
@@ -707,11 +794,89 @@ public:
             x = x*a + 1;
             return r;
         }
-        FastRandom( unsigned seed ) {
+        explicit FastRandom( unsigned seed ) {
             x = seed;
             a = Primes[seed % (sizeof(Primes) / sizeof(Primes[0]))];
         }
     };
+    template<typename T>
+    class FastRandomBody {
+        FastRandom r;
+    public:
+        explicit FastRandomBody( unsigned seed ) : r(seed) {}
+        // Depending on the input type T the result distribution formed from this operator()
+        // might possess different characteristics than the original one used in FastRandom instance.
+        T operator()() { return T(r.get()); }
+    };
+
+    int SetEnv( const char *envname, const char *envval ) {
+        ASSERT( envname && envval, "Harness::SetEnv() requires two valid C strings" );
+#if __TBB_WIN8UI_SUPPORT
+        ASSERT( false, "Harness::SetEnv() should not be called in code built for win8ui" );
+        return -1;
+#elif !(_MSC_VER || __MINGW32__ || __MINGW64__)
+        // On POSIX systems use setenv
+        return setenv(envname, envval, /*overwrite=*/1);
+#elif __STDC_SECURE_LIB__>=200411
+        // this macro is set in VC & MinGW if secure API functions are present
+        return _putenv_s(envname, envval);
+#else
+        // If no secure API on Windows, use _putenv
+        size_t namelen = strlen(envname), valuelen = strlen(envval);
+        char* buf = new char[namelen+valuelen+2];
+        strncpy(buf, envname, namelen);
+        buf[namelen] = '=';
+        strncpy(buf+namelen+1, envval, valuelen);
+        buf[namelen+1+valuelen] = char(0);
+        int status = _putenv(buf);
+        delete[] buf;
+        return status;
+#endif
+    }
+
+    char* GetEnv(const char *envname) {
+        ASSERT(envname, "Harness::GetEnv() requires a valid C string");
+#if __TBB_WIN8UI_SUPPORT
+        return NULL;
+#else
+        return std::getenv(envname);
+#endif
+    }
+
+    long GetIntEnv( const char * envname ) {
+        ASSERT(envname, "Harness::GetIntEnv() requires a valid C string");
+#if !__TBB_WIN8UI_SUPPORT
+        if( const char* s = std::getenv(envname) ){
+            char* end = NULL;
+            errno = 0;
+            long value = std::strtol(s, &end, 10);
+
+            // We have exceeded the range, value is negative or string is incovertable
+            if ( errno == ERANGE || value < 0 || end==s )
+            {
+                return -1;
+            }
+
+            for ( ; *end != '\0'; end++ )
+            {
+                if ( !std::isspace(*end) )
+                    return -1;
+            }
+
+            return value;
+        }
+#endif
+        return -1;
+    }
+
+    class DummyBody {
+        int m_numIters;
+    public:
+        explicit DummyBody( int iters ) : m_numIters( iters ) {}
+        void operator()( int ) const {
+            for ( volatile int i = 0; i < m_numIters; ++i ) {}
+        }
+    };
 } // namespace Harness
 
 #endif /* tbb_tests_harness_H */
