@@ -1,24 +1,20 @@
 /*
-    Copyright 2005-2015 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2020 Intel Corporation
 
-    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
-    you can redistribute it and/or modify it under the terms of the GNU General Public License
-    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
-    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
-    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-    See  the GNU General Public License for more details.   You should have received a copy of
-    the  GNU General Public License along with Threading Building Blocks; if not, write to the
-    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
 
-    As a special exception,  you may use this file  as part of a free software library without
-    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
-    functions from this file, or you compile this file and link it with other files to produce
-    an executable,  this file does not by itself cause the resulting executable to be covered
-    by the GNU General Public License. This exception does not however invalidate any other
-    reasons why the executable file might be covered by the GNU General Public License.
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
 */
 
-/** @file harness_graph.cpp     
+/** @file harness_graph.cpp
     This contains common helper classes and functions for testing graph nodes
 **/
 
@@ -26,14 +22,20 @@
 #define harness_graph_H
 
 #include "harness.h"
+#include "harness_barrier.h"
 #include "tbb/flow_graph.h"
 #include "tbb/null_rw_mutex.h"
 #include "tbb/atomic.h"
 #include "tbb/concurrent_unordered_map.h"
 #include "tbb/task.h"
 #include "tbb/task_scheduler_init.h"
+#include "tbb/compat/condition_variable"
+#include "tbb/mutex.h"
+#include "tbb/tbb_thread.h"
 
-#define WAIT_MAX 100000
+using tbb::flow::internal::SUCCESSFULLY_ENQUEUED;
+
+#define WAIT_MAX 2000000
 #define BACKOFF_WAIT(ex,msg) \
 { \
     int wait_cnt = 0; \
@@ -61,21 +63,21 @@
 // conversion operators to the class, since we don't want it in general,
 // only in these tests.
 template<typename InputType, typename OutputType>
-struct convertor {
+struct converter {
     static OutputType convert_value(const InputType &i) {
         return OutputType(i);
     }
 };
 
 template<typename InputType>
-struct convertor<InputType,tbb::flow::continue_msg> {
+struct converter<InputType,tbb::flow::continue_msg> {
     static tbb::flow::continue_msg convert_value(const InputType &/*i*/) {
-        return tbb::flow::continue_msg(); 
+        return tbb::flow::continue_msg();
     }
 };
 
 template<typename OutputType>
-struct convertor<tbb::flow::continue_msg,OutputType> {
+struct converter<tbb::flow::continue_msg,OutputType> {
     static OutputType convert_value(const tbb::flow::continue_msg &/*i*/) {
         return OutputType();
     }
@@ -86,7 +88,7 @@ template<size_t N>
 struct mof_helper {
     template<typename InputType, typename ports_type>
     static inline void output_converted_value(const InputType &i, ports_type &p) {
-        (void)tbb::flow::get<N-1>(p).try_put(convertor<InputType,typename tbb::flow::tuple_element<N-1,ports_type>::type::output_type>::convert_value(i));
+        (void)tbb::flow::get<N-1>(p).try_put(converter<InputType,typename tbb::flow::tuple_element<N-1,ports_type>::type::output_type>::convert_value(i));
         output_converted_value<N-1>(i, p);
     }
 };
@@ -96,7 +98,7 @@ struct mof_helper<1> {
     template<typename InputType, typename ports_type>
     static inline void output_converted_value(const InputType &i, ports_type &p) {
         // just emit a default-constructed object
-        (void)tbb::flow::get<0>(p).try_put(convertor<InputType,typename tbb::flow::tuple_element<0,ports_type>::type::output_type>::convert_value(i));
+        (void)tbb::flow::get<0>(p).try_put(converter<InputType,typename tbb::flow::tuple_element<0,ports_type>::type::output_type>::convert_value(i));
     }
 };
 
@@ -154,11 +156,11 @@ struct harness_graph_executor {
     static inline OutputType func( InputType v ) {
         size_t c; // Declaration separate from initialization to avoid ICC internal error on IA-64 architecture
         c = current_executors.fetch_and_increment();
-        ASSERT( max_executors == 0 || c <= max_executors, NULL ); 
+        ASSERT( max_executors == 0 || c <= max_executors, NULL );
         ++execute_count;
         OutputType v2 = (*fptr)(v);
         current_executors.fetch_and_decrement();
-        return v2; 
+        return v2;
     }
 
     template< typename RW >
@@ -177,7 +179,7 @@ struct harness_graph_executor {
         OutputType operator()( InputType i ) {
            typename RW::scoped_lock l( harness_graph_executor::mutex_holder<RW>::mutex, /*write=*/false );
            my_execute_count.fetch_and_increment();
-           return harness_graph_executor::func(i); 
+           return harness_graph_executor::func(i);
         }
     };
     typedef tfunctor<tbb::null_rw_mutex> functor;
@@ -200,11 +202,13 @@ struct harness_graph_multifunction_executor {
     static tbb::atomic<size_t> current_executors;
     static size_t max_executors;
 
+    static inline void empty_func( const InputType&, ports_type& ) {
+    }
 
     static inline void func( const InputType &v, ports_type &p ) {
         size_t c; // Declaration separate from initialization to avoid ICC internal error on IA-64 architecture
         c = current_executors.fetch_and_increment();
-        ASSERT( max_executors == 0 || c <= max_executors, NULL ); 
+        ASSERT( max_executors == 0 || c <= max_executors, NULL );
         ASSERT(tbb::flow::tuple_size<OutputTuple>::value == 1, NULL);
         ++execute_count;
         (*fptr)(v,p);
@@ -271,13 +275,14 @@ size_t harness_graph_multifunction_executor<InputType, OutputTuple>::max_executo
 
 //! Counts the number of puts received
 template< typename T >
-struct harness_counting_receiver : public tbb::flow::receiver<T>, NoCopy {
+struct harness_counting_receiver : public tbb::flow::receiver<T>, NoAssign {
 
     tbb::atomic< size_t > my_count;
     T max_value;
     size_t num_copies;
+    tbb::flow::graph& my_graph;
 
-    harness_counting_receiver() : num_copies(1) {
+    harness_counting_receiver(tbb::flow::graph& g) : num_copies(1), my_graph(g) {
        my_count = 0;
     }
 
@@ -287,9 +292,13 @@ struct harness_counting_receiver : public tbb::flow::receiver<T>, NoCopy {
        num_copies = c;
     }
 
-    /* override */ tbb::task *try_put_task( const T & ) {
+    tbb::flow::graph& graph_reference() const __TBB_override {
+        return my_graph;
+    }
+
+    tbb::task *try_put_task( const T & ) __TBB_override {
       ++my_count;
-      return const_cast<tbb::task *>(tbb::flow::interface7::SUCCESSFULLY_ENQUEUED);
+      return const_cast<tbb::task *>(SUCCESSFULLY_ENQUEUED);
     }
 
     void validate() {
@@ -297,21 +306,18 @@ struct harness_counting_receiver : public tbb::flow::receiver<T>, NoCopy {
         ASSERT( n == num_copies*max_value, NULL );
     }
 
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+#if TBB_DEPRECATED_FLOW_NODE_EXTRACTION
     typedef typename tbb::flow::receiver<T>::built_predecessors_type built_predecessors_type;
     built_predecessors_type mbp;
-    /*override*/ built_predecessors_type &built_predecessors() { return mbp; }
+    built_predecessors_type &built_predecessors() __TBB_override { return mbp; }
     typedef typename tbb::flow::receiver<T>::predecessor_list_type predecessor_list_type;
-    /*override*/void internal_add_built_predecessor(tbb::flow::sender<T> &) {}
-    /*override*/void internal_delete_built_predecessor(tbb::flow::sender<T> &) {}
-    /*override*/void copy_predecessors(predecessor_list_type &) { }
-    /*override*/size_t predecessor_count() { return 0; }
-    /*override*/void clear_predecessors() { my_count = 0; };
-    /*override*/void reset_receiver(tbb::flow::reset_flags /*f*/) { my_count = 0; }
-#else
-    /*override*/void reset_receiver() { my_count = 0; }
+    typedef typename tbb::flow::receiver<T>::predecessor_type predecessor_type;
+    void internal_add_built_predecessor(predecessor_type &) __TBB_override {}
+    void internal_delete_built_predecessor(predecessor_type &) __TBB_override {}
+    void copy_predecessors(predecessor_list_type &) __TBB_override { }
+    size_t predecessor_count() __TBB_override { return 0; }
 #endif
-
+    void reset_receiver(tbb::flow::reset_flags /*f*/) __TBB_override { my_count = 0; }
 };
 
 //! Counts the number of puts received
@@ -323,8 +329,9 @@ struct harness_mapped_receiver : public tbb::flow::receiver<T>, NoCopy {
     size_t num_copies;
     typedef tbb::concurrent_unordered_map< T, tbb::atomic< size_t > > map_type;
     map_type *my_map;
+    tbb::flow::graph& my_graph;
 
-    harness_mapped_receiver() : my_map(NULL) {
+    harness_mapped_receiver(tbb::flow::graph& g) : my_map(NULL), my_graph(g) {
        my_count = 0;
     }
 
@@ -340,7 +347,7 @@ struct harness_mapped_receiver : public tbb::flow::receiver<T>, NoCopy {
        my_map = new map_type;
     }
 
-    /* override */ tbb::task * try_put_task( const T &t ) {
+    tbb::task * try_put_task( const T &t ) __TBB_override {
       if ( my_map ) {
           tbb::atomic<size_t> a;
           a = 1;
@@ -352,7 +359,11 @@ struct harness_mapped_receiver : public tbb::flow::receiver<T>, NoCopy {
       } else {
           ++my_count;
       }
-      return const_cast<tbb::task *>(tbb::flow::interface7::SUCCESSFULLY_ENQUEUED);
+      return const_cast<tbb::task *>(SUCCESSFULLY_ENQUEUED);
+    }
+
+    tbb::flow::graph& graph_reference() const __TBB_override {
+        return my_graph;
     }
 
     void validate() {
@@ -366,20 +377,22 @@ struct harness_mapped_receiver : public tbb::flow::receiver<T>, NoCopy {
             ASSERT( n == num_copies*max_value, NULL );
         }
     }
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+#if TBB_DEPRECATED_FLOW_NODE_EXTRACTION
     typedef typename tbb::flow::receiver<T>::built_predecessors_type built_predecessors_type;
     built_predecessors_type mbp;
-    /*override*/ built_predecessors_type &built_predecessors() { return mbp; }
+    built_predecessors_type &built_predecessors() __TBB_override { return mbp; }
     typedef typename tbb::flow::receiver<T>::predecessor_list_type predecessor_list_type;
-    /*override*/void internal_add_built_predecessor(tbb::flow::sender<T> &) {}
-    /*override*/void internal_delete_built_predecessor(tbb::flow::sender<T> &) {}
-    /*override*/void copy_predecessors(predecessor_list_type &) { }
-    /*override*/size_t predecessor_count() { return 0; }
-    /*override*/void clear_predecessors() { my_count = 0; };
-    /*override*/void reset_receiver(tbb::flow::reset_flags /*f*/) { my_count = 0; if(my_map) delete my_map; my_map = new map_type; }
-#else
-    /*override*/void reset_receiver() { my_count = 0; if(my_map) delete my_map; my_map = new map_type; }
+    typedef typename tbb::flow::receiver<T>::predecessor_type predecessor_type;
+    void internal_add_built_predecessor(predecessor_type &) __TBB_override {}
+    void internal_delete_built_predecessor(predecessor_type &) __TBB_override {}
+    void copy_predecessors(predecessor_list_type &) __TBB_override { }
+    size_t predecessor_count() __TBB_override { return 0; }
 #endif
+    void reset_receiver(tbb::flow::reset_flags /*f*/) __TBB_override {
+        my_count = 0;
+        if(my_map) delete my_map;
+        my_map = new map_type;
+    }
 
 };
 
@@ -387,7 +400,7 @@ struct harness_mapped_receiver : public tbb::flow::receiver<T>, NoCopy {
 template< typename T >
 struct harness_counting_sender : public tbb::flow::sender<T>, NoCopy {
 
-    typedef tbb::flow::receiver<T> successor_type;
+    typedef typename tbb::flow::sender<T>::successor_type successor_type;
     tbb::atomic< successor_type * > my_receiver;
     tbb::atomic< size_t > my_count;
     tbb::atomic< size_t > my_received;
@@ -405,30 +418,29 @@ struct harness_counting_sender : public tbb::flow::sender<T>, NoCopy {
        my_received = 0;
     }
 
-    /* override */ bool register_successor( successor_type &r ) {
+    bool register_successor( successor_type &r ) __TBB_override {
         my_receiver = &r;
         return true;
     }
 
-    /* override */ bool remove_successor( successor_type &r ) {
+    bool remove_successor( successor_type &r ) __TBB_override {
         successor_type *s = my_receiver.fetch_and_store( NULL );
         ASSERT( s == &r, NULL );
         return true;
     }
 
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+#if TBB_DEPRECATED_FLOW_NODE_EXTRACTION
     typedef typename tbb::flow::sender<T>::successor_list_type successor_list_type;
     typedef typename tbb::flow::sender<T>::built_successors_type built_successors_type;
     built_successors_type bst;
-    /*override*/ built_successors_type &built_successors() { return bst; }
-    /* override */ void internal_add_built_successor( successor_type &) {}
-    /* override */ void internal_delete_built_successor( successor_type &) {}
-    /* override */ void copy_successors(successor_list_type &) { }
-    /*override*/ void clear_successors() { my_receiver = NULL; }
-    /* override */ size_t successor_count() { return 0; }
+    built_successors_type &built_successors() __TBB_override { return bst; }
+    void internal_add_built_successor( successor_type &) __TBB_override {}
+    void internal_delete_built_successor( successor_type &) __TBB_override {}
+    void copy_successors(successor_list_type &) __TBB_override { }
+    size_t successor_count() __TBB_override { return 0; }
 #endif
 
-    /* override */ bool try_get( T & v ) { 
+    bool try_get( T & v ) __TBB_override {
         size_t i = my_count.fetch_and_increment();
         if ( i < my_limit ) {
            v = T( i );
@@ -457,16 +469,16 @@ struct harness_counting_sender : public tbb::flow::sender<T>, NoCopy {
         while ( s->try_put( T(i) ) ) {
             ++my_received;
             i = my_count.fetch_and_increment();
-        } 
+        }
     }
 
     void try_put_until_limit() {
         successor_type *s = my_receiver;
 
-        for ( int i = 0; i < (int)my_limit; ++i ) { 
+        for ( int i = 0; i < (int)my_limit; ++i ) {
             ASSERT( s->try_put( T(i) ), NULL );
             ++my_received;
-        } 
+        }
         ASSERT( my_received == my_limit, NULL );
     }
 
@@ -512,7 +524,7 @@ struct serial_continue_body {
     }
 };
 
-#if TBB_PREVIEW_FLOW_GRAPH_FEATURES
+#if TBB_DEPRECATED_FLOW_NODE_EXTRACTION
 
 
 // walk two lists via iterator, match elements of each, in possibly-different ordder, and
@@ -540,6 +552,7 @@ found_it:
     }
     return true;
 }
+#endif  /* TBB_DEPRECATED_FLOW_NODE_EXTRACTION */
 
 template<typename T, typename BufferType>
 void test_resets() {
@@ -551,7 +564,7 @@ void test_resets() {
     tbb::flow::queue_node<T>  q0(g);
     T j;
     bool nFound[NN];
-    
+
     // reset empties buffer
     for(T i = 0; i < NN; ++i) {
         b0.try_put(i);
@@ -651,10 +664,12 @@ void test_resets() {
         g.wait_for_all();
         ASSERT((int)serial_fn_state0 == 1, "function_node executed when it shouldn't");
         T outt;
-        ASSERT(b0.try_get(outt) && (T)23 == outt, "node lost its input"); 
+        ASSERT(b0.try_get(outt) && (T)23 == outt, "node lost its input");
     }
 }
 
+#if TBB_DEPRECATED_FLOW_NODE_EXTRACTION
+
 template< typename NODE_TYPE >
 class test_buffer_base_extract {
 protected:
@@ -683,15 +698,15 @@ protected:
     typename NODE_TYPE::successor_list_type::iterator ms_list_iter;
 
     virtual void set_up_lists() {
-        in0_p_list.clear(); 
+        in0_p_list.clear();
         in0_s_list.clear();
-        in1_p_list.clear(); 
+        in1_p_list.clear();
         in1_s_list.clear();
-        mp_list.clear(); 
+        mp_list.clear();
         ms_list.clear();
-        out0_p_list.clear(); 
+        out0_p_list.clear();
         out0_s_list.clear();
-        out1_p_list.clear(); 
+        out1_p_list.clear();
         out1_s_list.clear();
         in0.copy_predecessors(in0_p_list);
         in0.copy_successors(in0_s_list);
@@ -732,20 +747,20 @@ protected:
         int first_pred = *(mp_list.begin()) == ins[0] ? 0 : ( *(mp_list.begin()) == ins[1] ? 1 : -1 );
         mp_list_iter = mp_list.begin(); ++mp_list_iter;
         int second_pred = *mp_list_iter == ins[0] ? 0 : ( *mp_list_iter == ins[1] ? 1 : -1 );
-        ASSERT( first_pred != -1 && second_pred != -1 && first_pred != second_pred, "bad predecessor(s) for middle" ); 
+        ASSERT( first_pred != -1 && second_pred != -1 && first_pred != second_pred, "bad predecessor(s) for middle" );
 
         int first_succ = *(ms_list.begin()) == outs[0] ? 0 : ( *(ms_list.begin()) == outs[1] ? 1 : -1 );
         ms_list_iter = ++(ms_list.begin());
         int second_succ = *ms_list_iter == outs[0] ? 0 : ( *ms_list_iter == outs[1] ? 1 : -1 );
-        ASSERT( first_succ != -1 && second_succ != -1 && first_succ != second_succ, "bad successor(s) for middle" ); 
- 
+        ASSERT( first_succ != -1 && second_succ != -1 && first_succ != second_succ, "bad successor(s) for middle" );
+
         in0.try_put(1);
         in1.try_put(2);
         g.wait_for_all();
 
         int r = 0;
         int v = 0;
-    
+
         ASSERT( in0.try_get(v) == false, "buffer should not have a value" );
         ASSERT( in1.try_get(v) == false, "buffer should not have a value" );
         ASSERT( middle.try_get(v) == false, "buffer should not have a value" );
@@ -781,17 +796,17 @@ protected:
         ASSERT( out0.successor_count() == 0 && out0_s_list.size() == 0, "expected 0 successors" );
         ASSERT( out1.predecessor_count() == 1 && out1_p_list.size() == 1 && *(out1_p_list.begin()) == mp_ptr, "expected 1 predecessor" );
         ASSERT( out1.successor_count() == 0 && out1_s_list.size() == 0, "expected 0 successors" );
-    
+
         ASSERT( middle.predecessor_count() == 1 && mp_list.size() == 1, "expected two predecessors" );
         ASSERT( middle.successor_count() == 1 && ms_list.size() == 1, "expected two successors" );
-    
+
         ASSERT( *(mp_list.begin()) == ins[1], "incorrect predecessor" );
         ASSERT( *(ms_list.begin()) == outs[1], "incorrect successor" );
-    
+
         in0.try_put(1);
         in1.try_put(2);
         g.wait_for_all();
-    
+
         int v = 0;
         ASSERT( in0.try_get(v) == true && v == 1, "buffer should have a value of 1" );
         ASSERT( in1.try_get(v) == false, "buffer should not have a value" );
@@ -819,14 +834,14 @@ protected:
         ASSERT( out0.successor_count() == 0 && out0_s_list.size() == 0, "expected 0 successors" );
         ASSERT( out1.predecessor_count() == 0 && out1_p_list.size() == 0, "expected 0 predecessors" );
         ASSERT( out1.successor_count() == 0 && out1_s_list.size() == 0, "expected 0 successors" );
-    
+
         ASSERT( middle.predecessor_count() == 0 && mp_list.size() == 0, "expected 0 predecessors" );
         ASSERT( middle.successor_count() == 0 && ms_list.size() == 0, "expected 0 successors" );
-    
+
         in0.try_put(1);
         in1.try_put(2);
         g.wait_for_all();
-    
+
         int v = 0;
         ASSERT( in0.try_get(v) == true && v == 1, "buffer should have a value of 1" );
         ASSERT( in1.try_get(v) == true && v == 2, "buffer should have a value of 2" );
@@ -841,7 +856,7 @@ protected:
 
 public:
 
-    test_buffer_base_extract(tbb::flow::graph &_g, NODE_TYPE &i0, NODE_TYPE &i1, NODE_TYPE &m, NODE_TYPE &o0, NODE_TYPE &o1) : 
+    test_buffer_base_extract(tbb::flow::graph &_g, NODE_TYPE &i0, NODE_TYPE &i1, NODE_TYPE &m, NODE_TYPE &o0, NODE_TYPE &o1) :
         g(_g), in0(i0), in1(i1), middle(m), out0(o0), out1(o1) {
         ins[0] = &in0;
         ins[1] = &in1;
@@ -850,7 +865,7 @@ public:
         ms_ptr = static_cast< typename NODE_TYPE::successor_type * >(&middle);
         mp_ptr = static_cast< typename NODE_TYPE::predecessor_type *>(&middle);
     }
- 
+
     virtual ~test_buffer_base_extract() {}
 
     void run_tests() {
@@ -871,7 +886,7 @@ public:
 
         make_and_validate_full_graph();
     }
-   
+
 };
 
 template< typename NODE_TYPE >
@@ -884,7 +899,7 @@ protected:
     NODE_TYPE my_out0;
     NODE_TYPE my_out1;
 public:
-    test_buffer_extract() : test_buffer_base_extract<NODE_TYPE>( my_g, my_in0, my_in1, my_middle, my_out0, my_out1), 
+    test_buffer_extract() : test_buffer_base_extract<NODE_TYPE>( my_g, my_in0, my_in1, my_middle, my_out0, my_out1),
                             my_in0(my_g), my_in1(my_g), my_middle(my_g), my_out0(my_g), my_out1(my_g) { }
 };
 
@@ -905,15 +920,15 @@ protected:
     count_t out1_count;
 
     struct always_zero { size_t operator()(int) { return 0; } };
-    struct always_inc { 
+    struct always_inc {
         count_t *c;
         always_inc(count_t &_c) : c(&_c) {}
-        size_t operator()(int) { 
+        size_t operator()(int) {
             return c->fetch_and_increment();
-        } 
+        }
     };
 
-    /*override*/void set_up_lists() {
+    void set_up_lists() __TBB_override {
         middle_count = 0;
         out0_count = 0;
         out1_count = 0;
@@ -923,9 +938,9 @@ protected:
 
 
 public:
-    test_buffer_extract() : test_buffer_base_extract<my_node_t>( my_g, my_in0, my_in1, my_middle, my_out0, my_out1), 
-                            my_in0(my_g, always_zero()), my_in1(my_g, always_zero()), my_middle(my_g, always_inc(middle_count)), 
-                            my_out0(my_g, always_inc(out0_count)), my_out1(my_g, always_inc(out1_count)) { 
+    test_buffer_extract() : test_buffer_base_extract<my_node_t>( my_g, my_in0, my_in1, my_middle, my_out0, my_out1),
+                            my_in0(my_g, always_zero()), my_in1(my_g, always_zero()), my_middle(my_g, always_inc(middle_count)),
+                            my_out0(my_g, always_inc(out0_count)), my_out1(my_g, always_inc(out1_count)) {
     }
 };
 
@@ -945,7 +960,7 @@ void test_extract_on_node() {
         tbb::flow::make_edge(node0, q2);
         q0.try_put(ItemType(i));
         g.wait_for_all();
-    
+
         /* q0               */
         /*   \              */
         /*    \             */
@@ -953,14 +968,14 @@ void test_extract_on_node() {
         /*    /             */
         /*   /              */
         /* q1               */
-    
+
         ASSERT(node0.predecessor_count() == 2 && q0.successor_count() == 1 && q1.successor_count() == 1, "bad predecessor count");
         ASSERT(node0.successor_count() == 1 && q2.predecessor_count() == 1, "bad successor count");
 
         ASSERT(q2.try_get(dont_care) && int(dont_care) == i, "item not forwarded");
         typename NType<ItemType>::successor_list_type sv, sv1;
         typename NType<ItemType>::predecessor_list_type pv, pv1;
-    
+
         pv1.push_back(&q0);
         pv1.push_back(&q1);
         sv1.push_back(&q2);
@@ -968,7 +983,7 @@ void test_extract_on_node() {
         node0.copy_successors(sv);
         ASSERT(lists_match(pv,pv1), "predecessor vector incorrect");
         ASSERT(lists_match(sv,sv1), "successor vector incorrect");
-    
+
         if(i == 0) {
             node0.extract();
         }
@@ -993,7 +1008,229 @@ void test_extract_on_node() {
     }
 }
 
-#endif  // TBB_PREVIEW_FLOW_GRAPH_FEATURES
+#endif  // TBB_DEPRECATED_FLOW_NODE_EXTRACTION
+
+template<typename NodeType>
+void test_input_ports_return_ref(NodeType& mip_node) {
+    typename NodeType::input_ports_type& input_ports1 = mip_node.input_ports();
+    typename NodeType::input_ports_type& input_ports2 = mip_node.input_ports();
+    ASSERT(&input_ports1 == &input_ports2, "input_ports() should return reference");
+}
+
+template<typename NodeType>
+void test_output_ports_return_ref(NodeType& mop_node) {
+    typename NodeType::output_ports_type& output_ports1 = mop_node.output_ports();
+    typename NodeType::output_ports_type& output_ports2 = mop_node.output_ports();
+    ASSERT(&output_ports1 == &output_ports2, "output_ports() should return reference");
+}
+
+template< template <typename> class ReservingNodeType, typename DataType, bool DoClear >
+class harness_reserving_body : NoAssign {
+    ReservingNodeType<DataType> &my_reserving_node;
+    tbb::flow::buffer_node<DataType> &my_buffer_node;
+public:
+    harness_reserving_body(ReservingNodeType<DataType> &reserving_node, tbb::flow::buffer_node<DataType> &bn) : my_reserving_node(reserving_node), my_buffer_node(bn) {}
+    void operator()(DataType i) const {
+        my_reserving_node.try_put(i);
+#if _MSC_VER && !__INTEL_COMPILER
+#pragma warning (push)
+#pragma warning (disable: 4127)  /* suppress conditional expression is constant */
 #endif
+        if (DoClear) {
+#if _MSC_VER && !__INTEL_COMPILER
+#pragma warning (pop)
+#endif
+            my_reserving_node.clear();
+        }
+        my_buffer_node.try_put(i);
+        my_reserving_node.try_put(i);
+    }
+};
 
+template< template <typename> class ReservingNodeType, typename DataType >
+void test_reserving_nodes() {
+    const size_t N = 300;
 
+    tbb::flow::graph g;
+
+    ReservingNodeType<DataType> reserving_n(g);
+
+    tbb::flow::buffer_node<DataType> buffering_n(g);
+    tbb::flow::join_node< tbb::flow::tuple<DataType, DataType>, tbb::flow::reserving > join_n(g);
+    harness_counting_receiver< tbb::flow::tuple<DataType, DataType> > end_receiver(g);
+
+    tbb::flow::make_edge(reserving_n, tbb::flow::input_port<0>(join_n));
+    tbb::flow::make_edge(buffering_n, tbb::flow::input_port<1>(join_n));
+    tbb::flow::make_edge(join_n, end_receiver);
+
+    NativeParallelFor(N, harness_reserving_body<ReservingNodeType, DataType, false>(reserving_n, buffering_n));
+    g.wait_for_all();
+
+    ASSERT(end_receiver.my_count == N, NULL);
+
+    // Should not hang
+    NativeParallelFor(N, harness_reserving_body<ReservingNodeType, DataType, true>(reserving_n, buffering_n));
+    g.wait_for_all();
+
+    ASSERT(end_receiver.my_count == 2 * N, NULL);
+}
+
+namespace lightweight_testing {
+
+typedef tbb::flow::tuple<int, int> output_tuple_type;
+
+template<typename NodeType>
+class native_loop_body : NoAssign {
+    NodeType& my_node;
+public:
+    native_loop_body(NodeType& node) : my_node(node) {}
+
+    void operator()(int) const {
+        tbb::tbb_thread::id this_id = tbb::this_tbb_thread::get_id();
+        my_node.try_put(this_id);
+    }
+};
+
+class concurrency_checker_body {
+public:
+    tbb::atomic<unsigned> my_body_count;
+
+    concurrency_checker_body() {
+        my_body_count = 0;
+    }
+
+    template<typename gateway_type>
+    void operator()(const tbb::tbb_thread::id& input, gateway_type&) {
+        increase_and_check(input);
+    }
+
+    output_tuple_type operator()(const tbb::tbb_thread::id& input) {
+        increase_and_check(input);
+        return output_tuple_type();
+    }
+
+private:
+    void increase_and_check(const tbb::tbb_thread::id& input) {
+        ++my_body_count;
+        tbb::tbb_thread::id body_thread_id = tbb::this_tbb_thread::get_id();
+        ASSERT(input == body_thread_id, "Body executed as not lightweight");
+    }
+};
+
+template<typename NodeType>
+void test_unlimited_lightweight_execution(unsigned N) {
+    tbb::flow::graph g;
+    NodeType node(g, tbb::flow::unlimited, concurrency_checker_body());
+
+    NativeParallelFor(N, native_loop_body<NodeType>(node));
+    g.wait_for_all();
+
+    concurrency_checker_body body = tbb::flow::copy_body<concurrency_checker_body>(node);
+    ASSERT(body.my_body_count == N, "Body needs to be executed N times");
+}
+
+// Using TBB implementation of condition variable
+// not to include std header, which has problems with old GCC
+using tbb::interface5::condition_variable;
+using tbb::interface5::unique_lock;
+
+tbb::mutex m;
+condition_variable lightweight_condition;
+bool work_submitted;
+bool lightweight_work_processed;
+
+template<typename NodeType>
+class native_loop_limited_body : NoAssign {
+    NodeType& my_node;
+    Harness::SpinBarrier& my_barrier;
+public:
+    native_loop_limited_body(NodeType& node, Harness::SpinBarrier& barrier):
+        my_node(node), my_barrier(barrier) {}
+    void operator()(int) const {
+        tbb::tbb_thread::id this_id = tbb::this_tbb_thread::get_id();
+        my_node.try_put(this_id);
+        if(!lightweight_work_processed) {
+            my_barrier.wait();
+            work_submitted = true;
+            lightweight_condition.notify_all();
+        }
+    }
+};
+
+struct condition_predicate {
+    bool operator()() {
+        return work_submitted;
+    }
+};
+
+class limited_lightweight_checker_body {
+public:
+    tbb::atomic<unsigned> my_body_count;
+    tbb::atomic<unsigned> my_lightweight_count;
+    tbb::atomic<unsigned> my_task_count;
+    limited_lightweight_checker_body() {
+        my_body_count = 0;
+        my_lightweight_count = 0;
+        my_task_count = 0;
+    }
+private:
+    void increase_and_check(const tbb::tbb_thread::id& /*input*/) {
+        ++my_body_count;
+        bool is_task = tbb::task::self().state() == tbb::task::executing;
+        if(is_task) {
+            ++my_task_count;
+        } else {
+            unique_lock<tbb::mutex> lock(m);
+            lightweight_condition.wait(lock, condition_predicate());
+            ++my_lightweight_count;
+            lightweight_work_processed = true;
+        }
+    }
+public:
+    template<typename gateway_type>
+    void operator()(const tbb::tbb_thread::id& input, gateway_type&) {
+        increase_and_check(input);
+    }
+    output_tuple_type operator()(const tbb::tbb_thread::id& input) {
+        increase_and_check(input);
+        return output_tuple_type();
+    }
+};
+
+template<typename NodeType>
+void test_limited_lightweight_execution(unsigned N, unsigned concurrency) {
+    ASSERT(concurrency != tbb::flow::unlimited,
+           "Test for limited concurrency cannot be called with unlimited concurrency argument");
+    tbb::flow::graph g;
+    NodeType node(g, concurrency, limited_lightweight_checker_body());
+    // Execute first body as lightweight, then wait for all other threads to fill internal buffer.
+    // Then unblock the lightweightd thread and check if other body executions are inside tbb task.
+    Harness::SpinBarrier barrier(N - concurrency);
+    NativeParallelFor(N, native_loop_limited_body<NodeType>(node, barrier));
+    g.wait_for_all();
+    limited_lightweight_checker_body body = tbb::flow::copy_body<limited_lightweight_checker_body>(node);
+    ASSERT(body.my_body_count == N, "Body needs to be executed N times");
+    ASSERT(body.my_lightweight_count == concurrency, "Body needs to be executed as lightweight once");
+    ASSERT(body.my_task_count == N - concurrency, "Body needs to be executed as not lightweight N - 1 times");
+    work_submitted = false;
+    lightweight_work_processed = false;
+}
+
+template<typename NodeType>
+void test_lightweight(unsigned N) {
+    test_unlimited_lightweight_execution<NodeType>(N);
+    test_limited_lightweight_execution<NodeType>(N, tbb::flow::serial);
+    test_limited_lightweight_execution<NodeType>(N, (std::min)(tbb::tbb_thread::hardware_concurrency() / 2, N/2));
+}
+
+template<template<typename, typename, typename, typename> class NodeType>
+void test(unsigned N) {
+    typedef tbb::tbb_thread::id input_type;
+    typedef tbb::cache_aligned_allocator<input_type> allocator_type;
+    typedef NodeType<input_type, output_tuple_type, tbb::flow::queueing_lightweight, allocator_type> node_type;
+    test_lightweight<node_type>(N);
+}
+
+}
+
+#endif
