@@ -1,31 +1,24 @@
 /*
-    Copyright 2005-2015 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2020 Intel Corporation
 
-    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
-    you can redistribute it and/or modify it under the terms of the GNU General Public License
-    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
-    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
-    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-    See  the GNU General Public License for more details.   You should have received a copy of
-    the  GNU General Public License along with Threading Building Blocks; if not, write to the
-    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
 
-    As a special exception,  you may use this file  as part of a free software library without
-    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
-    functions from this file, or you compile this file and link it with other files to produce
-    an executable,  this file does not by itself cause the resulting executable to be covered
-    by the GNU General Public License. This exception does not however invalidate any other
-    reasons why the executable file might be covered by the GNU General Public License.
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
 */
 
-#ifndef TBB_PREVIEW_FLOW_GRAPH_NODES
-    #define TBB_PREVIEW_FLOW_GRAPH_NODES 1
-#endif
-
 #include "harness.h"
-#if __TBB_PREVIEW_COMPOSITE_NODE 
+#if __TBB_FLOW_GRAPH_CPP11_FEATURES
 
 #include "tbb/flow_graph.h"
+#include "harness_graph.h"
 #include <tuple>
 #include <cmath>
 #include <vector>
@@ -41,17 +34,17 @@ class src_body{
     int finish;
     int step;
 public:
-    src_body(int f, int s) : start(1), finish(f), step(s) {} 
+    src_body(int f, int s) : start(1), finish(f), step(s) {}
     bool operator()(int &a) {
        a = start;
        if (start <= finish) {
            a = start;
            start+=step;
            return true;
-       } 
+       }
        else {
            return false;
-       };  
+       };
    }
 };
 
@@ -69,7 +62,7 @@ int operator()(int i){return i;}
 };
 
 template<int N, typename T1, typename T2>
-struct compare { 
+struct compare {
     static void compare_refs(T1 tuple1, T2 tuple2) {
     ASSERT( &tbb::flow::get<N>(tuple1) == &tbb::flow::get<N>(tuple2), "ports not set correctly");
     compare<N-1, T1, T2>::compare_refs(tuple1, tuple2);
@@ -77,7 +70,7 @@ struct compare {
 };
 
 template<typename T1, typename T2>
-struct compare<1, T1, T2> { 
+struct compare<1, T1, T2> {
     static void compare_refs(T1 tuple1, T2 tuple2) {
     ASSERT(&tbb::flow::get<0>(tuple1) == &tbb::flow::get<0>(tuple2), "port 0 not correctly set");
     }
@@ -86,10 +79,10 @@ struct compare<1, T1, T2> {
 void add_all_nodes (){
     tbb::flow::graph g;
 
-    typedef tbb::flow::tuple<tbb::flow::continue_msg, tbb::flow::tuple<int, int>, int, int, int, int, 
-                             int, int, int, int, int, int, int, int > InputTupleType; 
+    typedef tbb::flow::tuple<tbb::flow::continue_msg, tbb::flow::tuple<int, int>, int, int, int, int,
+                             int, int, int, int, int, int, int, int > InputTupleType;
 
-    typedef tbb::flow::tuple<tbb::flow::continue_msg, tbb::flow::tuple<int, int>, tbb::flow::tagged_msg<size_t, int, float>, 
+    typedef tbb::flow::tuple<tbb::flow::continue_msg, tbb::flow::tuple<int, int>, tbb::flow::tagged_msg<size_t, int, float>,
                              int, int, int, int, int, int, int, int, int, int, int, int >  OutputTupleType;
 
     typedef tbb::flow::tuple< > EmptyTupleType;
@@ -118,33 +111,28 @@ void add_all_nodes (){
     tbb::flow::overwrite_node<int> ovw(g);
     tbb::flow::sequencer_node<int> seq(g, seq_body());
 
-    #if !(__GNUC__==4 && __GNUC_MINOR__==4  && !defined(__clang__)) 
-    // upcasting of a tuple of reference from derived to base fails on gcc4.4 (and all icc in this environment)
-    // if std::tie is used to create the tuple of references
-    auto input_tuple = std::tie(ct, s, m_fxn, fxn, bc, tbb::flow::input_port<0>(j), lim, q, tbb::flow::input_port<0>(ind), 
+#if !__TBB_UPCAST_OF_TUPLE_OF_REF_BROKEN
+    auto input_tuple = std::tie(ct, s, m_fxn, fxn, bc, tbb::flow::input_port<0>(j), lim, q, tbb::flow::input_port<0>(ind),
                                 pq, ovw, wo, bf, seq);
-    auto output_tuple = std::tie(ct,j, ind, fxn, src, bc, tbb::flow::output_port<0>(s), lim, tbb::flow::output_port<0>(m_fxn), 
+    auto output_tuple = std::tie(ct,j, ind, fxn, src, bc, tbb::flow::output_port<0>(s), lim, tbb::flow::output_port<0>(m_fxn),
                                  q, pq, ovw, wo, bf, seq );
-    #else
-    input_output_type::input_ports_type input_tuple(ct, s, m_fxn, fxn, bc, tbb::flow::input_port<0>(j), lim, q, 
+#else
+    // upcasting from derived to base for a tuple of references created by std::tie
+    // fails on gcc 4.4 (and all icc in that environment)
+    input_output_type::input_ports_type input_tuple(ct, s, m_fxn, fxn, bc, tbb::flow::input_port<0>(j), lim, q,
                                                     tbb::flow::input_port<0>(ind), pq, ovw, wo, bf, seq);
 
     input_output_type::output_ports_type output_tuple(ct,j, ind, fxn, src, bc, tbb::flow::output_port<0>(s),
                                                       lim, tbb::flow::output_port<0>(m_fxn), q, pq, ovw, wo, bf, seq);
-    #endif
+#endif
 
-    //composite_node with both input_ports and output_ports 
+    //composite_node with both input_ports and output_ports
     input_output_type a_node(g);
-    a_node.set_external_ports(input_tuple, output_tuple); 
-   
-    bool a_visible_nodes_added = false;
-    a_visible_nodes_added = a_node.add_visible_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
-    ASSERT(a_visible_nodes_added == true, "not all nodes were added"); 
+    a_node.set_external_ports(input_tuple, output_tuple);
+
+    a_node.add_visible_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
+    a_node.add_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
 
-    bool a_nodes_added = false;
-    a_nodes_added = a_node.add_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
-    ASSERT(a_nodes_added == true, "not all nodes were added");
-    
     auto a_node_input_ports_ptr = a_node.input_ports();
     compare<NUM_INPUTS-1, decltype(a_node_input_ports_ptr), decltype(input_tuple)>::compare_refs(a_node_input_ports_ptr, input_tuple);
     ASSERT (NUM_INPUTS == tbb::flow::tuple_size<decltype(a_node_input_ports_ptr)>::value, "not all declared input ports were bound to nodes");
@@ -154,32 +142,23 @@ void add_all_nodes (){
     ASSERT(NUM_OUTPUTS == tbb::flow::tuple_size<decltype(a_node_output_ports_ptr)>::value, "not all declared output ports were bound to nodes");
 
     //composite_node with only input_ports
-    input_only_type b_node(g); 
+    input_only_type b_node(g);
     b_node.set_external_ports(input_tuple);
 
-    bool b_visible_nodes_added = false;
-    b_visible_nodes_added = b_node.add_visible_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
-    ASSERT(b_visible_nodes_added == true, "not all nodes were added");
-
-    bool b_nodes_added = false;
-    b_nodes_added = b_node.add_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
-    ASSERT(b_nodes_added == true, "not all nodes were added");
+    b_node.add_visible_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
+    b_node.add_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
 
     auto b_node_input_ports_ptr = b_node.input_ports();
     compare<NUM_INPUTS-1, decltype(b_node_input_ports_ptr), decltype(input_tuple)>::compare_refs(b_node_input_ports_ptr, input_tuple);
     ASSERT (NUM_INPUTS == tbb::flow::tuple_size<decltype(b_node_input_ports_ptr)>::value, "not all declared input ports were bound to nodes");
 
     //composite_node with only output_ports
-    output_only_type c_node(g); 
+    output_only_type c_node(g);
     c_node.set_external_ports(output_tuple);
 
-    bool c_visible_nodes_added = false;
-    c_visible_nodes_added = c_node.add_visible_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
-    ASSERT(c_visible_nodes_added == true, "not all nodes were added");
+    c_node.add_visible_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
 
-    bool c_nodes_added = false;
-    c_nodes_added = c_node.add_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
-    ASSERT(c_nodes_added == true, "not all nodes were added");
+    c_node.add_nodes(src, fxn, m_fxn, bc, lim, ind, s, ct, j, q, bf, pq, wo, ovw, seq);
 
     auto c_node_output_ports_ptr = c_node.output_ports();
     compare<NUM_OUTPUTS-1, decltype(c_node_output_ports_ptr), decltype(output_tuple)>::compare_refs(c_node_output_ports_ptr, output_tuple);
@@ -194,18 +173,16 @@ struct tiny_node : public tbb::flow::composite_node< tbb::flow::tuple< int >, tb
 public:
     tiny_node(tbb::flow::graph &g, bool hidden = false) : base_type(g), f1(g, tbb::flow::unlimited, passthru_body() ), f2(g, tbb::flow::unlimited, passthru_body() ) {
         tbb::flow::make_edge( f1, f2 );
-        
+
         tbb::flow::tuple<tbb::flow::function_node< int, int >& > input_tuple(f1);
         tbb::flow::tuple<tbb::flow::function_node< int, int >& > output_tuple(f2);
         base_type::set_external_ports( input_tuple, output_tuple );
-        bool nodes_added = false;
 
         if(hidden)
-            nodes_added  = base_type::add_nodes(f1, f2);
+            base_type::add_nodes(f1, f2);
         else
-            nodes_added = base_type::add_visible_nodes(f1, f2);
+            base_type::add_visible_nodes(f1, f2);
 
-        ASSERT(nodes_added == true, "nodes not added properly");
     }
 };
 
@@ -220,6 +197,9 @@ int test_tiny(bool hidden = false) {
     ASSERT(&tbb::flow::get<0>(t1.input_ports()) == &t1.f1, "f1 not bound to input port 0 in composite_node t1");
     ASSERT(&tbb::flow::get<0>(t1.output_ports()) == &t1.f2, "f2 not bound to output port 0 in composite_node t1");
 
+    test_input_ports_return_ref(t1);
+    test_output_ports_return_ref(t1);
+
     tiny_node t2(g, hidden);
     ASSERT(&tbb::flow::input_port<0>(t2) == &t2.f1, "f1 not bound to input port 0 in composite_node t2");
     ASSERT(&tbb::flow::output_port<0>(t2) == &t2.f2, "f2 not bound to output port 0 in composite_node t2");
@@ -236,7 +216,7 @@ int test_tiny(bool hidden = false) {
 
     int i, j =0;
     q.try_get(i);
-    ASSERT( i == 1, "item did not go through graph");  
+    ASSERT( i == 1, "item did not go through graph");
     q.try_get(j);
     ASSERT( !j, "unexpected item in graph");
     g.wait_for_all();
@@ -247,20 +227,20 @@ int test_tiny(bool hidden = false) {
 
     tbb::flow::make_edge( t1 , f3 );
     tbb::flow::make_edge(f3, q);
-   
+
     f0.try_put(2);
     g.wait_for_all();
 
     q.try_get(i);
-    ASSERT( i == 2, "item did not go through graph after removal of edge");  
+    ASSERT( i == 2, "item did not go through graph after removal of edge");
     q.try_get(j);
     ASSERT( !j, "unexpected item in graph after removal of edge");
 
-    return 0; 
+    return 0;
 }
 
 class adder_node : public tbb::flow::composite_node< tbb::flow::tuple< int, int >, tbb::flow::tuple< int > > {
-public: 
+public:
     tbb::flow::join_node< tbb::flow::tuple< int, int >, tbb::flow::queueing > j;
     tbb::flow::function_node< tbb::flow::tuple< int, int >, int > f;
 private:
@@ -275,25 +255,21 @@ private:
 public:
     adder_node(tbb::flow::graph &g, bool hidden = false) : base_type(g), j(g), f(g, tbb::flow::unlimited, f_body() ) {
         tbb::flow::make_edge( j, f );
-        
-        base_type::input_ports_type input_tuple(tbb::flow::input_port<0>(j), tbb::flow::input_port<1>(j) );  
-        base_type::output_ports_type output_tuple(f);
-        base_type::set_external_ports( input_tuple, output_tuple );
-        bool nodes_added = false;
-       
-        if (hidden)
-           nodes_added = base_type::add_nodes(j, f);
-        else
-           nodes_added = base_type::add_visible_nodes(j, f);
 
-        ASSERT(nodes_added == true, "either j or f not added"); 
+        base_type::set_external_ports(base_type::input_ports_type(tbb::flow::input_port<0>(j), tbb::flow::input_port<1>(j)), base_type::output_ports_type(f));
+
+        if (hidden)
+            base_type::add_nodes(j, f);
+        else
+            base_type::add_visible_nodes(j, f);
+
     }
 };
 
 struct square_body { int operator()(int v) { return v*v; } };
 struct cube_body { int operator()(int v) { return v*v*v; } };
 int adder_sum(int i) {
-    return (int)(pow(3*pow(i,3) + pow(i, 2),2)); 
+    return (int)(pow(3*pow(i,3) + pow(i, 2),2));
 }
 int test_adder(bool hidden = false) {
     tbb::flow::graph g;
@@ -323,7 +299,7 @@ int test_adder(bool hidden = false) {
 
     tbb::flow::function_node<int,int> s2(g, tbb::flow::unlimited, square_body());
     tbb::flow::queue_node<int> q(g);
- 
+
     tbb::flow::make_edge( s, tbb::flow::input_port<0>(a0) );
     tbb::flow::make_edge( c, tbb::flow::input_port<1>(a0) );
 
@@ -335,7 +311,7 @@ int test_adder(bool hidden = false) {
 
     tbb::flow::make_edge( tbb::flow::output_port<0>(a2), s2 );
     tbb::flow::make_edge( s2, q );
-      
+
     int sum_total=0;
     int result=0;
     for ( int i = 1; i < 4; ++i ) {
@@ -348,7 +324,7 @@ int test_adder(bool hidden = false) {
     int j;
     for ( int i = 1; i < 4; ++i ) {
         q.try_get(j);
-        result += j;  
+        result += j;
     }
     g.wait_for_all();
     ASSERT(result == sum_total, "the sum from the graph does not match the calculated value");
@@ -371,12 +347,72 @@ int test_adder(bool hidden = false) {
 
     for ( int i = 10; i < 20; ++i ) {
         q.try_get(j);
-        result += j;  
+        result += j;
     }
     g.wait_for_all();
     ASSERT(result == sum_total, "the new sum after the replacement of the nodes does not match the calculated value");
 
-    return 0; 
+    return 0;
+}
+
+/*
+                                              outer composite node (outer_node)
+                                     |-------------------------------------------------------------------|
+                                     |                                                                   |
+                                     |  |------------------|  |------------------|  |------------------| |
+             |---------------------| |--| inner composite  | /| inner composite  | /| inner composite  | | |-------------------|
+             |broadcast node(input)|/|  | node             |/ | node             |/ | node             |-+-| queue node(output)|
+             |---------------------|\|  |(inner_node1)     |\ | (inner_node2)    |\ | (inner_node3)    | | |-------------------|
+                                     |--|                  | \|                  | \|                  | |
+                                     |  |------------------|  |------------------|  |------------------| |
+                                     |                                                                   |
+                                     |-------------------------------------------------------------------|
+
+*/
+int test_nested_adder(bool hidden=false) {
+    tbb::flow::graph g;
+    tbb::flow::composite_node<tbb::flow::tuple<int, int>, tbb::flow::tuple<int> > outer_node(g);
+    typedef tbb::flow::composite_node<tbb::flow::tuple<int, int>, tbb::flow::tuple<int> > base_type;
+    tbb::flow::broadcast_node<int> input(g);
+    tbb::flow::queue_node<int> output(g);
+
+    adder_node inner_node1(g, hidden);
+    adder_node inner_node2(g, hidden);
+    adder_node inner_node3(g, hidden);
+
+    outer_node.set_external_ports(base_type::input_ports_type(tbb::flow::input_port<0>(inner_node1), tbb::flow::input_port<1>(inner_node1)), base_type::output_ports_type(tbb::flow::output_port<0>(inner_node3)));
+
+    ASSERT(&tbb::flow::input_port<0>(outer_node) == &tbb::flow::input_port<0>(inner_node1), "input port 0 of inner_node1 not bound to input port 0 in outer_node");
+    ASSERT(&tbb::flow::input_port<1>(outer_node) == &tbb::flow::input_port<1>(inner_node1), "input port 1 of inner_node1 not bound to input port 1 in outer_node");
+    ASSERT(&tbb::flow::output_port<0>(outer_node) == &tbb::flow::output_port<0>(inner_node3), "output port 0 of inner_node3 not bound to output port 0 in outer_node");
+
+    tbb::flow::make_edge(input, tbb::flow::input_port<0>(outer_node)/*inner_node1*/);
+    tbb::flow::make_edge(input, tbb::flow::input_port<1>(outer_node)/*inner_node1*/);
+
+    tbb::flow::make_edge(inner_node1, tbb::flow::input_port<0>(inner_node2));
+    tbb::flow::make_edge(inner_node1, tbb::flow::input_port<1>(inner_node2));
+
+    tbb::flow::make_edge(inner_node2, tbb::flow::input_port<0>(inner_node3));
+    tbb::flow::make_edge(inner_node2, tbb::flow::input_port<1>(inner_node3));
+
+    tbb::flow::make_edge(outer_node/*inner_node3*/, output);
+
+    if(hidden)
+        outer_node.add_nodes(inner_node1, inner_node2, inner_node3);
+    else
+        outer_node.add_visible_nodes(inner_node1, inner_node2, inner_node3);
+
+    int out;
+    for (int i = 1; i < 200000; ++i) {
+        input.try_put(i);
+        g.wait_for_all();
+        output.try_get(out);
+        ASSERT(tbb::flow::output_port<0>(outer_node).try_get(out) == output.try_get(out), "output from outer_node does not match output from graph");
+        ASSERT(out == 8*i, "output from outer_node not correct");
+    }
+    g.wait_for_all();
+
+    return 0;
 }
 
 template< typename T >
@@ -403,7 +439,6 @@ public:
     prefix_node(tbb::flow::graph &g, bool hidden = false ) : base_type(g), j(g), s(g), f(g, tbb::flow::serial, f_body() ) {
         tbb::flow::make_edge( j, f );
         tbb::flow::make_edge( f, s );
-        bool nodes_added = false;
 
     typename base_type::input_ports_type input_tuple(tbb::flow::input_port<0>(j), tbb::flow::input_port<1>(j), tbb::flow::input_port<2>(j), tbb::flow::input_port<3>(j), tbb::flow::input_port<4>(j));
 
@@ -412,11 +447,10 @@ public:
     base_type::set_external_ports(input_tuple, output_tuple);
 
         if(hidden)
-            nodes_added = base_type::add_nodes(j,s,f);
+            base_type::add_nodes(j,s,f);
         else
-            nodes_added = base_type::add_visible_nodes(j,s,f);
-        
-        ASSERT(nodes_added == true, "nodes not properly added");
+            base_type::add_visible_nodes(j,s,f);
+
     }
 };
 
@@ -429,7 +463,7 @@ int test_prefix(bool hidden = false) {
     ASSERT(&tbb::flow::get<2>(p.input_ports()) == &tbb::flow::input_port<2>(p.j), "input port 2 of j is not bound to input port 2 of composite node p");
     ASSERT(&tbb::flow::input_port<3>(p.j) == &tbb::flow::input_port<3>(p.j), "input port 3 of j is not bound to input port 3 of composite node p");
     ASSERT(&tbb::flow::get<4>(p.input_ports()) == &tbb::flow::input_port<4>(p.j), "input port 4 of j is not bound to input port 4 of composite node p");
-    
+
 
     ASSERT(&tbb::flow::get<0>(p.output_ports()) == &tbb::flow::output_port<0>(p.s), "output port 0 of s is not bound to output port 0 of composite node p");
     ASSERT(&tbb::flow::output_port<1>(p.s) == &tbb::flow::output_port<1>(p.s), "output port 1 of s is not bound to output port 1 of composite node p");
@@ -461,23 +495,30 @@ int test_prefix(bool hidden = false) {
             g.wait_for_all();
         }
     }
-    return 0; 
+    return 0;
 }
 
+struct input_only_output_only_seq {
+    int operator()(int i){ return (i + 3) / 4 - 1;}
+};
+
 void input_only_output_only_composite(bool hidden) {
     tbb::flow::graph g;
-    tbb::flow::composite_node<tbb::flow::tuple<int>, tbb::flow::tuple<int> > input_output(g, "aaaaa");
+#if TBB_PREVIEW_FLOW_GRAPH_TRACE
+    tbb::flow::composite_node<tbb::flow::tuple<int>, tbb::flow::tuple<int> > input_output(g, "test_name");
+#else
+    tbb::flow::composite_node<tbb::flow::tuple<int>, tbb::flow::tuple<int> > input_output(g);
+#endif
     typedef tbb::flow::composite_node<tbb::flow::tuple<int>, tbb::flow::tuple<> > input_only_composite;
     typedef tbb::flow::composite_node<tbb::flow::tuple<>, tbb::flow::tuple<int> > output_only_composite;
     typedef tbb::flow::source_node<int> src_type;
     typedef tbb::flow::queue_node<int> q_type;
     typedef tbb::flow::function_node<int, int> f_type;
+    typedef tbb::flow::sequencer_node<int> sequencer_type;
 
     int num = 0;
     int finish=1000;
     int step = 4;
-    bool a_in_nodes_added = false;
-    bool a_out_nodes_added = false;
 
     input_only_composite a_in(g);
     output_only_composite a_out(g);
@@ -486,27 +527,29 @@ void input_only_output_only_composite(bool hidden) {
     q_type que(g);
     f_type f(g, 1, passthru_body());
 
+    // Sequencer_node is needed, because serial function_node guarantees only serial body execution,
+    // not a sequential order of messages dispatch
+    sequencer_type seq(g, input_only_output_only_seq());
+
     tbb::flow::tuple<f_type& > input_tuple(f);
     a_in.set_external_ports(input_tuple);
     ASSERT(&tbb::flow::get<0>(a_in.input_ports()) == &f, "f not bound to input port 0 in composite_node a_in");
 
-    tbb::flow::tuple<src_type&> output_tuple(src); 
+    tbb::flow::tuple<src_type&> output_tuple(src);
     a_out.set_external_ports(output_tuple);
     ASSERT(&tbb::flow::get<0>(a_out.output_ports()) == &src, "src not bound to output port 0 in composite_node a_out");
-    
-    if(hidden) { 
-        a_in_nodes_added = a_in.add_nodes(f, que); 
-        a_out_nodes_added = a_out.add_nodes(src); 
-    } else {
-        a_in_nodes_added = a_in.add_visible_nodes(f, que); 
-        a_out_nodes_added = a_out.add_visible_nodes(src);
 
-        ASSERT(a_in_nodes_added == true, "either f or que was not properly added to a_in");
-        ASSERT(a_out_nodes_added == true, "src node not properly added to a_out");  
-    }    
+    if(hidden) {
+        a_in.add_nodes(f, seq, que);
+        a_out.add_nodes(src);
+    } else {
+        a_in.add_visible_nodes(f, seq, que);
+        a_out.add_visible_nodes(src);
+    }
 
     tbb::flow::make_edge(a_out, a_in);
-    tbb::flow::make_edge(f, que);
+    tbb::flow::make_edge(f, seq);
+    tbb::flow::make_edge(seq, que);
     src.activate();
     g.wait_for_all();
 
@@ -517,25 +560,27 @@ void input_only_output_only_composite(bool hidden) {
     g.wait_for_all();
 }
 
-#endif // __TBB_PREVIEW_COMPOSITE_NODE
+#endif // __TBB_FLOW_GRAPH_CPP11_FEATURES
 
 int TestMain() {
 
-#if __TBB_PREVIEW_COMPOSITE_NODE 
+#if __TBB_FLOW_GRAPH_CPP11_FEATURES
 
     add_all_nodes();
     test_tiny(false);
     test_tiny(true);
     test_adder(false);
     test_adder(true);
+    test_nested_adder(true);
+    test_nested_adder(false);
     test_prefix(false);
     test_prefix(true);
     input_only_output_only_composite(true);
     input_only_output_only_composite(false);
 
-    return Harness::Done; 
-#else 
-    return Harness::Skipped; 
-#endif 
+    return Harness::Done;
+#else
+    return Harness::Skipped;
+#endif
 
 }
